#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pygame
from vec2d import vec2d
from itertools import cycle
import levels
import random
import levels


class Mutation(object):
    kind = None
    image = pygame.Surface((9, 9))
    image.set_colorkey((0, 0, 0))
    def attack(self, position, user, level):
        pass
    def mutate(self, person):
        pass
    def demutate(self, person):
        pass

class FasterMutation(Mutation):
    def __init__(self):
        self.image = pygame.image.load("miniFasterMutation.png").convert_alpha()
    kind = "moving"
    def mutate(self, person):
        person.acceleration += 0.5
    def demutate(self, person):
        person.acceleration -= 0.5

class HigherMutation(Mutation):
    def __init__(self):
        self.image = pygame.image.load("miniHigherMutation.png").convert_alpha()
    kind = "moving"
    def mutate(self, person):
        person.jumpPower += vec2d(0, -1)
    def demutate(self, person):
        person.jumpPower -= vec2d(0, -1)

class DoubleJumpMutation(Mutation):
    def __init__(self):
        self.image = pygame.image.load("miniDoubleJumpMutation.png").convert_alpha()
    kind = "special"
    def mutate(self, person):
        person.maxJumps += 1
    def demutate(self, person):
        person.maxJumps -= 1

class TripleJumpMutation(Mutation):
    def __init__(self):
        self.image = pygame.image.load("miniTripleJumpMutation.png").convert_alpha()
    kind = "special"
    def mutate(self, person):
        person.maxJumps += 2
    def demutate(self, person):
        person.maxJumps -= 2

class SpikeMutation(Mutation):
    def __init__(self):
        self.image = pygame.image.load("miniSpikeMutation.png").convert_alpha()
    kind = "attack"
    def attack(self, position, user, level):
        level.objects["etc"].add(Spikes(position, user))

class BombMutation(Mutation):
    def __init__(self):
        self.image = pygame.image.load("miniBombMutation.png").convert_alpha()
    kind = "attack"
    def attack(self, position, user, level):
        if not any(x.user==user for x in level.objects["bombs"].sprites()):
            level.objects["bombs"].add(Bomb(position, user))






class GameObject(pygame.sprite.Sprite):
    def __init__(self, position):
        pygame.sprite.Sprite.__init__(self)
        
        self.position = vec2d(position)
        self.image = pygame.Surface((10, 10))
    
    def get_rect(self):
        return self.image.get_rect(center=self.position)
    rect = property(get_rect)
    
    def update(self, level, game):
        pass


class Person(GameObject):
    def __init__(self, position):
        GameObject.__init__(self, position)
        
        self.speed = vec2d(0, 0)
        self.maxSpeed = 8
        self.uMaxSpeed = 20
        self.acceleration = 2
        self.friction = 0.8
        self.jumpPower = vec2d(0, -4)
        self.jumpRepeat = 8
        self.jumps = 1
        self.maxJumps = 1
        self.jumpCounter = 5
        self.jumping = False
        self.gravity = vec2d(0, 1.5)
        self.health = 3
        
        self.state = "walk"
        self.aniDelay = 30
        self.animation = {"walk": cycle(map( lambda x: x.convert_alpha(x), 
                                map(pygame.image.load, ["walk1.png", "walk2.png", "walk3.png", "walk2.png"]))), 
                          "stand": cycle([pygame.image.load("walk2.png").convert_alpha()])}
        self.currentFrame = self.animation[self.state].next()
        self.direction = False
        
        self.attackMutation = Mutation()
        self.movingMutation = Mutation()
        self.specialMutation = Mutation()
        
        self.recoil = 30
        
        self.stunned = 30
        
        self.hitSound = pygame.mixer.Sound("hit.wav")
        self.jumpSound = pygame.mixer.Sound("jump.wav")
        self.shootSound = pygame.mixer.Sound("shoot.wav")
        
    def get_image(self):
        if self.aniDelay:
            self.aniDelay -= 1
        else:
            self.aniDelay = 30
            self.currentFrame = self.animation[self.state].next()
        if self.speed.x < 0:
            self.direction = True
        elif self.speed.x > 0:
            self.direction = False
        return pygame.transform.flip(self.currentFrame, self.direction, False)
    def set_image(self, image): pass
    image = property(get_image, set_image)
    
    def add_mutation(self, mutation, kind):
        if kind == "moving":
            self.movingMutation.demutate(self)
            self.movingMutation = mutation
            self.movingMutation.mutate(self)
        elif kind == "special":
            self.specialMutation.demutate(self)
            self.specialMutation = mutation
            self.specialMutation.mutate(self)
        elif kind == "attack":
            self.attackMutation.demutate(self)
            self.attackMutation = mutation
            self.attackMutation.mutate(self)
    
    def take_damage(self):
        if not self.stunned:
            self.hitSound.play()
            self.health -= 1
            self.stunned = 60
    
    def get_actions(self, level):
        pass
        
    def update(self, level, game):
        
        actions = self.get_actions(level)
        
        if self.jumpCounter: self.jumpCounter -= 1
        
        if "jump" in actions:
            if level.tileMap.hit_test(self.rect.midbottom + vec2d(0, 2)):
                self.jumps = self.maxJumps
                self.jumpCounter = 0
            if self.jumps > 0 and not self.jumpCounter:
                self.jumpSound.play()
                self.jumping = True
                self.jumps -= 1
            self.jumpCounter = 5
        else:
            self.jumpCounter = 0
        
        if self.jumping and self.jumpRepeat:
            self.speed += self.jumpPower
            self.jumpRepeat -= 1
        else:
            self.jumping = False
            self.jumpRepeat = 8
        
        if self.recoil:
            self.recoil -= 1
        
        if "shoot" in actions and not self.recoil:
            self.shootSound.play()
            keys = pygame.key.get_pressed()
            direction = vec2d(0, 0)
            if keys[pygame.K_RIGHT]: direction += vec2d(1, 0)
            if keys[pygame.K_LEFT]: direction += vec2d(-1, 0)
            if keys[pygame.K_UP]: direction += vec2d(0, -1)
            if keys[pygame.K_DOWN]: direction += vec2d(0, 1)
            if direction == vec2d(0, 0): direction = vec2d(-1, 0) if self.direction else vec2d(1, 0)
            direction = direction.normalized()
            level.objects["bullets"].add(MutationBullet(self.position, direction))
            self.recoil = 30
        
        if self.stunned:
            self.stunned -= 1
        
        if "attack" in actions:
            self.attackMutation.attack(self.position, self, level)
        
        self.state = "stand"
        
        if "right" in actions:
            self.state = "walk"
            if abs(self.speed.x) < self.maxSpeed:
                self.speed.x += self.acceleration
        elif "left" in actions:
            self.state = "walk"
            if abs(self.speed.x) < self.maxSpeed:
                self.speed.x -= self.acceleration
        
        self.speed += self.gravity
        
        while level.tileMap.hit_test(self.rect.midtop+self.speed+vec2d(0, 2)):
            self.speed += vec2d(0, 1)
        while level.tileMap.hit_test(self.rect.midright+self.speed):
            self.speed += vec2d(-1, 0)
        while level.tileMap.hit_test(self.rect.midleft+self.speed):
            self.speed += vec2d(1, 0)
        while level.tileMap.hit_test(self.rect.midbottom+self.speed+vec2d(0, -2)):
            self.speed += vec2d(0, -1)
      
        self.speed *= self.friction
        
        if self.speed.length > self.uMaxSpeed:
            self.speed.length = self.uMaxSpeed
                
        
        self.speed.x = int(self.speed.x)
        self.speed.y = int(self.speed.y)
        self.position += self.speed
        














class Player(Person):
    
    def __init__(self, position):
        Person.__init__(self, position)
        
    def update(self, level, game):
        Person.update(self, level, game)
        if not self.health:
            game.level = levels.load_level(game.currentLevel)

    def get_actions(self, level):
        
        keys = pygame.key.get_pressed()
        actions = []
        
        if keys[pygame.K_s]:
            actions.append("jump")
        
        if keys[pygame.K_d]:
            actions.append("shoot")
        
        if keys[pygame.K_a]:
            return ["attack"]
        
        if keys[pygame.K_RIGHT] and keys[pygame.K_LEFT]:
            pass
        elif keys[pygame.K_RIGHT]:
            actions.append("right") 
        elif keys[pygame.K_LEFT]:
            actions.append("left") 
        
        return actions










class Enemy(Person):
    
    def __init__(self, position, movingMutation=Mutation(), specialMutation=Mutation(), attackMutation=Mutation()):
        Person.__init__(self, position)
        
        self.add_mutation(movingMutation, "moving")
        self.add_mutation(specialMutation, "special")
        self.add_mutation(attackMutation, "attack")
        
        self.maxJumps = 1
        self.health = 1
        
        self.direc = vec2d(1, 0)
        self.attack = 0
        self.animation = {"walk": cycle(map( lambda x: x.convert_alpha(x), 
                                map(pygame.image.load, ["ewalk1.png", "ewalk2.png", "ewalk3.png", "ewalk2.png"]))), 
                          "stand": cycle([pygame.image.load("ewalk2.png").convert_alpha()])}

    
    def update(self, level, game):
        Person.update(self, level, game)
        if not self.health:
            level.objects["enemies"].remove(self)
    
    def get_actions(self, level):
        actions = []
        
        if level.tileMap.hit_test(self.position + self.direc * 9):
            self.direc = -self.direc
        if self.direc == vec2d(1, 0):
            actions.append("right")
        if self.direc == vec2d(-1, 0):
            actions.append("left")
        
        if random.randrange(120) == 3:
            actions.append("jump")

        
        if self.attack:
            self.attack -= 1
            return ["attack"]
        else:
            if random.randrange(120) == 3:
                self.attack = 10

        return actions           


class CleverEnemy(Enemy):
    def get_actions(self, level):
        actions = []
        
        self.direc = vec2d((level.objects["player"].sprite.position - self.position).x, 0)
        self.direc = self.direc.normalized()
        
        if len(level.objects["bombs"].sprites()):
            if (self.position - min(map(lambda x: x.position, level.objects["bombs"].sprites()))).length < 20:
                self.direc = vec2d((self.position - min(map(lambda x: x.position,level.objects["bombs"].sprites()))).x, 0).normalized()
                actions.append("jump")
            
        if level.tileMap.hit_test(self.position + self.direc * 9):
            self.direc = -self.direc
        if self.direc == vec2d(1, 0):
            actions.append("right")
        if self.direc == vec2d(-1, 0):
            actions.append("left")
        
        if self.maxJumps:
            if self.speed.length == 0:
                actions.append("jump")
        
        if self.attackMutation != Mutation():
            if not (self.attackMutation == BombMutation() and any(x.user is self for x in level.objects["bombs"].sprites())):
                if self.position.get_distance(level.objects["player"].sprite.position) < 10:
                    return ["attack"]
        
       
        return actions           



class MutationBullet(GameObject):
    def __init__(self, position, direction):
        GameObject.__init__(self, position)
        self.image = pygame.image.load("bullet.png").convert_alpha()
        self.direction = direction
        self.age = 0
        
    def update(self, level, game):
        self.age += 1
        if self.age > 100:
            level.objects["bullets"].remove(self)
        self.position += self.direction * 3
        enemy = pygame.sprite.spritecollide(self, level.objects["enemies"], False)
        if len(enemy):
            enemy = enemy[0]
        else:
            return
        
        mov = enemy.movingMutation
        spe = enemy.specialMutation
        att = enemy.attackMutation
        enemy.add_mutation(level.objects["player"].sprite.movingMutation, "moving")
        enemy.add_mutation(level.objects["player"].sprite.specialMutation, "special")
        enemy.add_mutation(level.objects["player"].sprite.attackMutation, "attack")
        level.objects["player"].sprite.add_mutation(mov, "moving")
        level.objects["player"].sprite.add_mutation(spe, "special")
        level.objects["player"].sprite.add_mutation(att, "attack")
        level.objects["bullets"].remove(self)




class Spikes(GameObject):
    def __init__(self, position, user):
        GameObject.__init__(self, position)
        self.live = 2
        self.user = user
        self.image = pygame.image.load("spikes.png").convert_alpha()
    def update(self, level, game):
        colliders = pygame.sprite.spritecollide(self, level.objects["enemies"], False) + \
                    pygame.sprite.spritecollide(self, level.objects["player"], False)
        
        for collider in (x for x in colliders if isinstance(x, Person) and not (x is self.user)):
            collider.take_damage()
        
        self.live -= 1
        if not self.live:
            level.objects["etc"].remove(self)

class Bomb(GameObject):
    def __init__(self, position, user):
        GameObject.__init__(self, position)
        self.live = 40
        self.user = user
        self.image = pygame.image.load("bomb.png").convert_alpha()
        self.explosion = pygame.image.load("explosion.png").convert_alpha()
        self.sound = pygame.mixer.Sound("explosion.wav")
    def update(self, level, game):
        if self.live < 10:
            self.image = self.explosion
            self.sound.play()
            colliders = pygame.sprite.spritecollide(self, level.objects["enemies"], False) + \
                        pygame.sprite.spritecollide(self, level.objects["player"], False)
            
            for collider in colliders:
                collider.take_damage()
        
        self.position += vec2d(0, 2)
        while level.tileMap.hit_test(self.position+vec2d(0, 3)):
            self.position.y -= 1
        
        self.live -= 1
        if not self.live:
            level.objects["bombs"].remove(self)


class HUD(GameObject):
    def __init__(self, position):
        GameObject.__init__(self, position)
        self.image = pygame.Surface((300, 200))
        self.image.fill((255, 0, 0))
        self.image.set_colorkey((255, 0, 0))
        self.mutationScreen = pygame.image.load("mutationScreen.png").convert_alpha()
        self.heart = pygame.image.load("heart.png").convert_alpha()
    
    def update(self, level, game):
        self.position = level.objects["player"].sprite.position
        
        self.image = pygame.Surface((300, 200))
        self.image.fill((255, 0, 0))
        self.image.set_colorkey((255, 0, 0))
        
        
        mov = level.objects["player"].sprite.movingMutation.image
        spe = level.objects["player"].sprite.specialMutation.image
        att = level.objects["player"].sprite.attackMutation.image
        
        offset = vec2d(20, 20)
        self.image.blit(mov, offset + vec2d(0, 9))
        self.image.blit(spe, offset + vec2d(10, 9))
        self.image.blit(att, offset + vec2d(5, 1))
        
        self.image.blit(self.mutationScreen, offset)
        
        health = level.objects["player"].sprite.health
        for x in range(health):
            self.image.blit(self.heart, vec2d(13, 40) + vec2d(x * 12, 0))


class Exit(GameObject):
    def __init__(self, position, levelName):
        GameObject.__init__(self, position)
        self.image = pygame.image.load("exit.png")
        self.levelName = levelName
    def update(self, level, game):
        if pygame.sprite.spritecollide(self, level.objects["player"], False):
            game.currentLevel = self.levelName
            nLevel = levels.load_level(self.levelName)
            player = level.objects["player"].sprite
            player.position = vec2d(nLevel.objects["player"].sprite.position)
            player.speed = vec2d(0, 0)
            nLevel.objects["player"].add(player)
            game.level = nLevel



class Image(GameObject):
    def __init__(self, position, imagef):
        GameObject.__init__(self, position)
        self.image = pygame.image.load(imagef)
















